"""redo-always: tell redo that the current target is always out of date."""
import sys, os
from . import env, logs, state


def main():
    try:
        env.inherit()
        logs.setup(
            tty=sys.stderr, parent_logs=env.v.LOG,
            pretty=env.v.PRETTY, color=env.v.COLOR)

        me = os.path.join(env.v.STARTDIR,
                          os.path.join(env.v.PWD, env.v.TARGET))
        f = state.File(name=me)
        f.add_dep('m', state.ALWAYS)
        always = state.File(name=state.ALWAYS)
        always.stamp = state.STAMP_MISSING
        always.set_changed()
        always.save()
        state.commit()
    except KeyboardInterrupt:
        sys.exit(200)


if __name__ == '__main__':
    main()



"""Some helper functions that don't fit anywhere else."""
import os, errno, fcntl


class ImmediateReturn(Exception):
    def __init__(self, rv):
        Exception.__init__(self, "immediate return with exit code %d" % rv)
        self.rv = rv


def unlink(f):
    """Delete a file at path 'f' if it currently exists.

    Unlike os.unlink(), does not throw an exception if the file didn't already
    exist.
    """
    try:
        os.unlink(f)
    except OSError as e:
        if e.errno == errno.ENOENT:
            pass  # it doesn't exist, that's what you asked for


def close_on_exec(fd, yes):
    fl = fcntl.fcntl(fd, fcntl.F_GETFD)
    fl &= ~fcntl.FD_CLOEXEC
    if yes:
        fl |= fcntl.FD_CLOEXEC
    fcntl.fcntl(fd, fcntl.F_SETFD, fl)


def fd_exists(fd):
    try:
        fcntl.fcntl(fd, fcntl.F_GETFD)
    except IOError:
        return False
    return True



"""Code for manipulating redo's state database."""
import sys, os, errno, stat, fcntl, sqlite3
from . import cycles, env
from .helpers import unlink, close_on_exec
from .logs import warn, debug2, debug3

SCHEMA_VER = 2
TIMEOUT = 60

ALWAYS = '//ALWAYS'   # an invalid filename that is always marked as dirty
STAMP_DIR = 'dir'     # the stamp of a directory; mtime is unhelpful
STAMP_MISSING = '0'   # the stamp of a nonexistent file

LOG_LOCK_MAGIC = 0x10000000  # fid offset for "log locks"


def _connect(dbfile):
    _db = sqlite3.connect(dbfile, timeout=TIMEOUT)
    _db.execute("pragma synchronous = off")
    # Some old/broken versions of pysqlite on MacOS work badly with journal
    # mode PERSIST.  But WAL fails on Windows WSL due to WSL's totally broken
    # locking.  On WSL, at least PERSIST works in single-threaded mode, so
    # if we're careful we can use it, more or less.
    jmode = 'PERSIST' if env.v.LOCKS_BROKEN else 'WAL'
    _db.execute("pragma journal_mode = %s" % (jmode,))
    _db.text_factory = str
    return _db


# We need to keep a process-wide fd open for all access to the lock file.
# Because POSIX lock files are insane, if you close *one* fd pointing
# at a given inode, it will immediately release *all* locks on that inode from
# your pid, even if those locks are on a different fd.  This is literally
# never what you want.  To avoid the problem, always use just a single fd.
_lockfile = None


_db = None
def db():
    """Initialize the state database and return its object."""
    global _db, _lockfile
    if _db:
        return _db

    dbdir = '%s/.redo' % env.v.BASE
    dbfile = '%s/db.sqlite3' % dbdir
    try:
        os.mkdir(dbdir)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass  # if it exists, that's okay
        else:
            raise

    _lockfile = os.open(os.path.join(env.v.BASE, '.redo/locks'),
                        os.O_RDWR | os.O_CREAT, 0o666)
    close_on_exec(_lockfile, True)
    if env.is_toplevel and detect_broken_locks():
        env.mark_locks_broken()

    must_create = not os.path.exists(dbfile)
    if not must_create:
        _db = _connect(dbfile)
        try:
            row = _db.cursor().execute("select version from Schema").fetchone()
        except sqlite3.OperationalError:
            row = None
        ver = row and row[0] or None
        if ver != SCHEMA_VER:
            # Don't use err() here because this might happen before
            # redo-log spawns.
            sys.stderr.write(
                'redo: %s: found v%s (expected v%s)\n'
                % (dbfile, ver, SCHEMA_VER))
            sys.stderr.write('redo: manually delete .redo dir to start over.\n')
            sys.exit(1)
    if must_create:
        unlink(dbfile)
        _db = _connect(dbfile)
        _db.execute("create table Schema "
                    "    (version int)")
        _db.execute("create table Runid "
                    "    (id integer primary key autoincrement)")
        _db.execute("create table Files "
                    "    (name not null primary key, "
                    "     is_generated int, "
                    "     is_override int, "
                    "     checked_runid int, "
                    "     changed_runid int, "
                    "     failed_runid int, "
                    "     stamp, "
                    "     csum)")
        _db.execute("create table Deps "
                    "    (target int, "
                    "     source int, "
                    "     mode not null, "
                    "     delete_me int, "
                    "     primary key (target,source))")
        _db.execute("insert into Schema (version) values (?)", [SCHEMA_VER])
        # eat the '0' runid and File id.
        # Because of the cheesy way t/flush-cache is implemented, leave a
        # lot of runids available before the "first" one so that we
        # can adjust cached values to be before the first value.
        _db.execute("insert into Runid values (1000000000)")
        _db.execute("insert into Files (name) values (?)", [ALWAYS])

    if not env.v.RUNID:
        _db.execute("insert into Runid values "
                    "     ((select max(id)+1 from Runid))")
        env.v.RUNID = _db.execute("select last_insert_rowid()").fetchone()[0]
        os.environ['REDO_RUNID'] = str(env.v.RUNID)

    _db.commit()
    return _db


def init(targets):
    env.init(targets)
    db()
    if env.is_toplevel and detect_broken_locks():
        env.mark_locks_broken()


_wrote = 0
def _write(q, l):
    if _insane:
        return
    global _wrote
    _wrote += 1
    db().execute(q, l)


def commit():
    if _insane:
        return
    global _wrote
    if _wrote:
        db().commit()
        _wrote = 0


def rollback():
    if _insane:
        return
    global _wrote
    if _wrote:
        db().rollback()
        _wrote = 0


def is_flushed():
    return not _wrote


_insane = None
def check_sane():
    global _insane
    if not _insane:
        _insane = not os.path.exists('%s/.redo' % env.v.BASE)
    return not _insane


def _realdirpath(t):
    """Like realpath(), but don't follow symlinks for the last element.

    redo needs this because targets can be symlinks themselves, and we want
    to talk about the symlink, not what it points at.  However, all the path
    elements along the way could result in pathname aliases for a *particular*
    target, so we want to resolve it to one unique name.
    """
    dname, fname = os.path.split(t)
    if dname:
        dname = os.path.realpath(dname)
    return os.path.join(dname, fname)


_cwd = None
def relpath(t, base):
    """Given a relative or absolute path t, express it relative to base."""
    global _cwd
    if not _cwd:
        _cwd = os.getcwd()
    t = os.path.normpath(_realdirpath(os.path.join(_cwd, t)))
    base = os.path.normpath(_realdirpath(base))
    tparts = t.split('/')
    bparts = base.split('/')

    # zip must not return an iterator in python 3, because the source lists of
    # the iterators are changed. The iterator does not notice that and ends too
    # soon.
    for tp, bp in list(zip(tparts, bparts)):
        if tp != bp:
            break
        tparts.pop(0)
        bparts.pop(0)
    while bparts:
        tparts.insert(0, '..')
        bparts.pop(0)
    return '/'.join(tparts)


# Return a relative path for t that will work after we do
# chdir(dirname(env.v.TARGET)).
#
# This is tricky!  STARTDIR+PWD is the directory for the *dofile*, when
# the dofile was started.  However, inside the dofile, someone may have done
# a chdir to anywhere else.  env.v.TARGET is relative to the dofile path, so
# we have to first figure out where the dofile was, then find TARGET relative
# to that, then find t relative to that.
#
# FIXME: find some cleaner terminology for all these different paths.
def target_relpath(t):
    dofile_dir = os.path.abspath(os.path.join(env.v.STARTDIR, env.v.PWD))
    target_dir = os.path.abspath(
        os.path.dirname(os.path.join(dofile_dir, env.v.TARGET)))
    return relpath(t, target_dir)


def detect_override(stamp1, stamp2):
    """Determine if two stamps differ in a way that means manual override.

    When two stamps differ at all, that means the source is dirty and so we
    need to rebuild.  If they differ in mtime or size, then someone has surely
    edited the file, and we don't want to trample their changes.

    But if the only difference is something else (like ownership, st_mode,
    etc) then that might be a false positive; it's annoying to mark as
    overridden in that case, so we return False.  (It's still dirty though!)
    """
    if stamp1 == stamp2:
        return False
    crit1 = stamp1.split('-', 2)[0:2]
    crit2 = stamp2.split('-', 2)[0:2]
    return crit1 != crit2


def warn_override(name):
    warn('%s - you modified it; skipping\n' % name)


_file_cols = ['rowid', 'name', 'is_generated', 'is_override',
              'checked_runid', 'changed_runid', 'failed_runid',
              'stamp', 'csum']
class File(object):
    """An object representing a source or target in the redo database."""

    # use this mostly to avoid accidentally assigning to typos
    __slots__ = ['id'] + _file_cols[1:]

    # These warnings are a result of the weird way this class is
    # initialized, which we should fix, and then re-enable warning.
    # pylint: disable=attribute-defined-outside-init
    def _init_from_idname(self, fid, name, allow_add):
        q = ('select %s from Files ' % ', '.join(_file_cols))
        if fid != None:
            q += 'where rowid=?'
            l = [fid]
        elif name != None:
            name = (name == ALWAYS) and ALWAYS or relpath(name, env.v.BASE)
            q += 'where name=?'
            l = [name]
        else:
            raise Exception('name or id must be set')
        d = db()
        row = d.execute(q, l).fetchone()
        if not row:
            if not name:
                raise KeyError('No file with id=%r name=%r' % (fid, name))
            elif not allow_add:
                raise KeyError('No file with name=%r' % (name,))
            try:
                _write('insert into Files (name) values (?)', [name])
            except sqlite3.IntegrityError:
                # some parallel redo probably added it at the same time; no
                # big deal.
                pass
            row = d.execute(q, l).fetchone()
            assert row
        return self._init_from_cols(row)

    def _init_from_cols(self, cols):
        (self.id, self.name, self.is_generated, self.is_override,
         self.checked_runid, self.changed_runid, self.failed_runid,
         self.stamp, self.csum) = cols
        if self.name == ALWAYS and (
            self.changed_runid is None or self.changed_runid < env.v.RUNID):
            self.changed_runid = env.v.RUNID

    def __init__(self, fid=None, name=None, cols=None, allow_add=True):
        if cols:
            self._init_from_cols(cols)
        else:
            self._init_from_idname(fid, name, allow_add=allow_add)

    def __repr__(self):
        return "File(%r)" % (self.nicename(),)

    def refresh(self):
        self._init_from_idname(self.id, None, allow_add=False)

    def save(self):
        cols = ', '.join(['%s=?'%i for i in _file_cols[2:]])
        _write('update Files set '
               '    %s '
               '    where rowid=?' % cols,
               [self.is_generated, self.is_override,
                self.checked_runid, self.changed_runid, self.failed_runid,
                self.stamp, self.csum,
                self.id])

    def set_checked(self):
        self.checked_runid = env.v.RUNID

    def set_checked_save(self):
        self.set_checked()
        self.save()

    def set_changed(self):
        debug2('BUILT: %r (%r)\n' % (self.name, self.stamp))
        self.changed_runid = env.v.RUNID
        self.failed_runid = None
        self.is_override = False

    def set_failed(self):
        debug2('FAILED: %r\n' % self.name)
        self.update_stamp()
        self.failed_runid = env.v.RUNID
        if self.stamp != STAMP_MISSING:
            # if we failed and the target file still exists,
            # then we're generated.
            self.is_generated = True
        else:
            # if the target file now does *not* exist, then go back to
            # treating this as a source file.  Since it doesn't exist,
            # if someone tries to rebuild it immediately, it'll go
            # back to being a target.  But if the file is manually
            # created before that, we don't need a "manual override"
            # warning.
            self.is_generated = False

    def set_static(self):
        self.update_stamp(must_exist=True)
        self.failed_runid = None
        self.is_override = False
        self.is_generated = False

    def set_override(self):
        self.update_stamp()
        self.failed_runid = None
        self.is_override = True

    def update_stamp(self, must_exist=False):
        newstamp = self.read_stamp()
        if must_exist and newstamp == STAMP_MISSING:
            raise Exception("%r does not exist" % self.name)
        if newstamp != self.stamp:
            debug2("STAMP: %s: %r -> %r\n" % (self.name, self.stamp, newstamp))
            self.stamp = newstamp
            self.set_changed()

    def is_source(self):
        """Returns true if this object represents a source (not a target)."""
        if self.name.startswith('//'):
            return False  # special name, ignore
        newstamp = self.read_stamp()
        if (self.is_generated and
                (not self.is_failed() or newstamp != STAMP_MISSING) and
                not self.is_override and
                self.stamp == newstamp):
            # target is as we left it
            return False
        if ((not self.is_generated or self.stamp != newstamp) and
                newstamp == STAMP_MISSING):
            # target has gone missing after the last build.
            # It's not usefully a source *or* a target.
            return False
        return True

    def is_target(self):
        """Returns true if this object represents a target (not a source)."""
        if not self.is_generated:
            return False
        if self.is_source():
            return False
        return True

    def is_checked(self):
        return self.checked_runid and self.checked_runid >= env.v.RUNID

    def is_changed(self):
        return self.changed_runid and self.changed_runid >= env.v.RUNID

    def is_failed(self):
        return self.failed_runid and self.failed_runid >= env.v.RUNID

    def deps(self):
        """Return the list of objects that this object depends on."""
        if self.is_override or not self.is_generated:
            return
        q = ('select Deps.mode, Deps.source, %s '
             '  from Files '
             '    join Deps on Files.rowid = Deps.source '
             '  where target=?' % ', '.join(_file_cols[1:]))
        for row in db().execute(q, [self.id]).fetchall():
            mode = row[0]
            cols = row[1:]
            assert mode in ('c', 'm')
            yield mode, File(cols=cols)

    def zap_deps1(self):
        """Mark the list of dependencies of this object as deprecated.

        We do this when starting a new build of the current target.  We don't
        delete them right away, because if the build fails, we still want to
        know the old deps.
        """
        debug2('zap-deps1: %r\n' % self.name)
        _write('update Deps set delete_me=? where target=?', [True, self.id])

    def zap_deps2(self):
        """Delete any deps that were *not* referenced in the current run.

        Dependencies of a given target can change from one build to the next.
        We forget old dependencies only after a build completes successfully.
        """
        debug2('zap-deps2: %r\n' % self.name)
        _write('delete from Deps where target=? and delete_me=1', [self.id])

    def add_dep(self, mode, dep):
        src = File(name=dep)
        debug3('add-dep: "%s" < %s "%s"\n' % (self.name, mode, src.name))
        assert self.id != src.id
        _write("insert or replace into Deps "
               "    (target, mode, source, delete_me) values (?,?,?,?)",
               [self.id, mode, src.id, False])

    def _read_stamp_st(self, statfunc):
        try:
            st = statfunc(os.path.join(env.v.BASE, self.name))
        except OSError:
            return False, STAMP_MISSING
        if stat.S_ISDIR(st.st_mode):
            # directories change too much; detect only existence.
            return False, STAMP_DIR
        else:
            # a "unique identifier" stamp for a regular file
            return (
                stat.S_ISLNK(st.st_mode),
                '-'.join(str(s) for s in
                         ('%.6f' % st.st_mtime, st.st_size, st.st_ino,
                          st.st_mode, st.st_uid, st.st_gid))
            )

    def read_stamp(self):
        is_link, pre = self._read_stamp_st(os.lstat)
        if is_link:
            # if we're a symlink, we actually care about the link object
            # itself, *and* the target of the link.  If either changes,
            # we're considered dirty.
            #
            # On the other hand, detect_override() doesn't care about the
            # target of the link, only the link itself.
            _, post = self._read_stamp_st(os.stat)
            return pre + '+' + post
        else:
            return pre

    def nicename(self):
        return relpath(os.path.join(env.v.BASE, self.name), env.v.STARTDIR)


def files():
    q = ('select %s from Files order by name' % ', '.join(_file_cols))
    for cols in db().execute(q).fetchall():
        yield File(cols=cols)


def logname(fid):
    """Given the id of a File, return the filename of its build log."""
    return os.path.join(env.v.BASE, '.redo', 'log.%d' % fid)


# FIXME: I really want to use fcntl F_SETLK, F_SETLKW, etc here.  But python
# doesn't do the lockdata structure in a portable way, so we have to use
# fcntl.lockf() instead.  Usually this is just a wrapper for fcntl, so it's
# ok, but it doesn't have F_GETLK, so we can't report which pid owns the lock.
# The makes debugging a bit harder.  When we someday port to C, we can do that.
_locks = {}
class Lock(object):
    """An object representing a lock on a redo target file."""

    def __init__(self, fid):
        """Initialize a lock, given the target's state.File.id."""
        self.owned = False
        self.fid = fid
        assert _lockfile >= 0
        assert _locks.get(fid, 0) == 0
        _locks[fid] = 1

    def __del__(self):
        _locks[self.fid] = 0
        if self.owned:
            self.unlock()

    def check(self):
        """Check that this lock is in a sane state."""
        assert not self.owned
        cycles.check(self.fid)

    def trylock(self):
        """Non-blocking try to acquire our lock; returns true if it worked."""
        self.check()
        assert not self.owned
        try:
            fcntl.lockf(_lockfile, fcntl.LOCK_EX|fcntl.LOCK_NB, 1, self.fid)
        except IOError as e:
            if e.errno in (errno.EAGAIN, errno.EACCES):
                pass  # someone else has it locked
            else:
                raise
        else:
            self.owned = True
        return self.owned

    def waitlock(self, shared=False):
        """Try to acquire our lock, and wait if it's currently locked.

        If shared=True, acquires a shared lock (which can be shared with
        other shared locks; used by redo-log).  Otherwise, acquires an
        exclusive lock.
        """
        self.check()
        assert not self.owned
        fcntl.lockf(
            _lockfile,
            fcntl.LOCK_SH if shared else fcntl.LOCK_EX,
            1, self.fid)
        self.owned = True

    def unlock(self):
        """Release the lock, which we must currently own."""
        if not self.owned:
            raise Exception("can't unlock %r - we don't own it"
                            % self.fid)
        fcntl.lockf(_lockfile, fcntl.LOCK_UN, 1, self.fid)
        self.owned = False


def detect_broken_locks():
    """Detect Windows WSL's completely broken fcntl() locks.

    Symptom: locking a file always returns success, even if other processes
    also think they have it locked. See
    https://github.com/Microsoft/WSL/issues/1927 for more details.

    Bug exists at least in WSL "4.4.0-17134-Microsoft #471-Microsoft".

    Returns true if broken, false otherwise.
    """
    pl = Lock(0)
    # We wait for the lock here, just in case others are doing
    # this test at the same time.
    pl.waitlock(shared=False)
    pid = os.fork()
    if pid:
        # parent
        _, rv = os.waitpid(pid, 0)
        ok = os.WIFEXITED(rv) and not os.WEXITSTATUS(rv)
        return not ok
    else:
        # child
        try:
            # Doesn't actually unlock, since child process doesn't own it
            pl.unlock()
            del pl
            cl = Lock(0)
            # parent is holding lock, which should prevent us from getting it.
            owned = cl.trylock()
            if owned:
                # Got the lock? Yikes, the locking system is broken!
                os._exit(1)
            else:
                # Failed to get the lock? Good, the parent owns it.
                os._exit(0)
        except Exception:  # pylint: disable=broad-except
            import traceback
            traceback.print_exc()
        finally:
            os._exit(99)



"""Manage redo-related environment variables."""
import os, sys
from .atoi import atoi

is_toplevel = False

v = None


def _get(name, default):
    return os.environ.get(name, default)


def _get_int(name, default):
    return atoi(_get(name, str(default)))


def _get_bool(name, default):
    return 1 if _get(name, default) else 0


class Env(object):
    def __init__(self):
        # Mandatory.  If this is missing, you forgot to call init().
        self.BASE = os.environ['REDO_BASE'].rstrip('/')

        # Everything else, we can recover from defaults if unset.
        self.PWD = _get('REDO_PWD', '')
        self.TARGET = _get('REDO_TARGET', '')
        self.DEPTH = _get('REDO_DEPTH', '')
        self.DEBUG = atoi(_get('REDO_DEBUG', ''))
        self.DEBUG_LOCKS = _get_bool('REDO_DEBUG_LOCKS', '')
        self.DEBUG_PIDS = _get_bool('REDO_DEBUG_PIDS', '')
        self.LOCKS_BROKEN = _get_bool('REDO_LOCKS_BROKEN', '')
        self.VERBOSE = _get_int('REDO_VERBOSE', '')
        self.XTRACE = _get_int('REDO_XTRACE', '')
        self.KEEP_GOING = _get_bool('REDO_KEEP_GOING', '')
        self.LOG = _get_int('REDO_LOG', '1')
        self.LOG_INODE = _get('REDO_LOG_INODE', '')
        self.COLOR = _get_int('REDO_COLOR', '')
        self.PRETTY = _get_int('REDO_PRETTY', '')
        self.SHUFFLE = _get_bool('REDO_SHUFFLE', '')
        self.STARTDIR = _get('REDO_STARTDIR', '')
        self.RUNID = _get_int('REDO_RUNID', '') or None
        self.UNLOCKED = _get_bool('REDO_UNLOCKED', '')
        self.NO_OOB = _get_bool('REDO_NO_OOB', '')


def inherit():
    """Read environment (which must already be set) to get runtime settings."""

    if not os.environ.get('REDO'):
        sys.stderr.write('%s: error: must be run from inside a .do\n'
                         % sys.argv[0])
        sys.exit(100)

    global v
    v = Env()

    # not inheritable by subprocesses
    os.environ['REDO_UNLOCKED'] = ''
    os.environ['REDO_NO_OOB'] = ''


def init_no_state():
    """Start a session (if needed) for a command that needs no state db."""
    global is_toplevel
    if not os.environ.get('REDO'):
        os.environ['REDO'] = 'NOT_DEFINED'
        is_toplevel = True
    if not os.environ.get('REDO_BASE'):
        os.environ['REDO_BASE'] = 'NOT_DEFINED'
    inherit()


def init(targets):
    """Start a session (if needed) for a command that does need the state db.

    Args:
      targets: a list of targets we're trying to build.  We use this to
        help in guessing where the .redo database is located.
    """
    global is_toplevel
    if not os.environ.get('REDO'):
        # toplevel call to redo
        is_toplevel = True
        exenames = [os.path.abspath(sys.argv[0]),
                    os.path.realpath(sys.argv[0])]
        dirnames = [os.path.dirname(p) for p in exenames]
        trynames = ([os.path.abspath(p+'/../lib/redo') for p in dirnames] +
                    [p+'/../redo' for p in dirnames] +
                    dirnames)
        seen = {}
        dirs = []
        for k in trynames:
            if not seen.get(k):
                seen[k] = 1
                dirs.append(k)
        os.environ['PATH'] = ':'.join(dirs) + ':' + os.environ['PATH']
        os.environ['REDO'] = os.path.abspath(sys.argv[0])

    if not os.environ.get('REDO_BASE'):
        if not targets:
            # if no other targets given, assume the current directory
            targets = ['all']
        base = os.path.commonprefix([os.path.abspath(os.path.dirname(t))
                                     for t in targets] + [os.getcwd()])
        bsplit = base.split('/')
        for i in range(len(bsplit)-1, 0, -1):
            newbase = '/'.join(bsplit[:i])
            if os.path.exists(newbase + '/.redo'):
                base = newbase
                break
        os.environ['REDO_BASE'] = base
        os.environ['REDO_STARTDIR'] = os.getcwd()

    inherit()


def mark_locks_broken():
    """If file locking is broken, update the environment accordingly."""
    os.environ['REDO_LOCKS_BROKEN'] = '1'
    # FIXME: redo-log doesn't work when fcntl locks are broken.
    # We can probably work around that someday.
    os.environ['REDO_LOG'] = '0'
    inherit()



"""redo-ifcreate: build the current target if these targets are created."""
import sys, os
from . import env, logs, state
from .logs import err


def main():
    try:
        env.inherit()
        logs.setup(
            tty=sys.stderr, parent_logs=env.v.LOG,
            pretty=env.v.PRETTY, color=env.v.COLOR)

        me = os.path.join(env.v.STARTDIR,
                          os.path.join(env.v.PWD, env.v.TARGET))
        f = state.File(name=me)
        for t in sys.argv[1:]:
            if not t:
                err('cannot build the empty target ("").\n')
                sys.exit(204)
            if os.path.exists(t):
                err('redo-ifcreate: error: %r already exists\n' % t)
                sys.exit(1)
            else:
                f.add_dep('c', t)
        state.commit()
    except KeyboardInterrupt:
        sys.exit(200)


if __name__ == '__main__':
    main()



"""redo-unlocked: internal tool for building dependencies."""
import sys, os
from . import env, logs, state


def main():
    if len(sys.argv[1:]) < 2:
        sys.stderr.write('%s: at least 2 arguments expected.\n' % sys.argv[0])
        sys.exit(1)

    env.inherit()
    logs.setup(
        tty=sys.stderr, parent_logs=env.v.LOG,
        pretty=env.v.PRETTY, color=env.v.COLOR)

    target = sys.argv[1]
    deps = sys.argv[2:]

    for d in deps:
        assert d != target

    me = state.File(name=target)

    # Build the known dependencies of our primary target.  This *does* require
    # grabbing locks.
    os.environ['REDO_NO_OOB'] = '1'
    argv = ['redo-ifchange'] + deps
    rv = os.spawnvp(os.P_WAIT, argv[0], argv)
    if rv:
        sys.exit(rv)

    # We know our caller already owns the lock on target, so we don't have to
    # acquire another one; tell redo-ifchange about that.  Also, REDO_NO_OOB
    # persists from up above, because we don't want to do OOB now either.
    # (Actually it's most important for the primary target, since it's the one
    # who initiated the OOB in the first place.)
    os.environ['REDO_UNLOCKED'] = '1'
    argv = ['redo-ifchange', target]
    rv = os.spawnvp(os.P_WAIT, argv[0], argv)
    if rv:
        sys.exit(rv)


if __name__ == '__main__':
    main()



"""redo-ifchange: build the given targets if they have changed."""
import os, sys, traceback
from . import env, builder, deps, helpers, jobserver, logs, state
from .logs import debug2, err


def should_build(t):
    f = state.File(name=t)
    if f.is_failed():
        raise helpers.ImmediateReturn(32)
    dirty = deps.isdirty(f, depth='', max_changed=env.v.RUNID,
                         already_checked=[])
    return f.is_generated, dirty == [f] and deps.DIRTY or dirty


def main():
    rv = 202
    try:
        targets = sys.argv[1:]
        state.init(targets)
        if env.is_toplevel and not targets:
            targets = ['all']
        if env.is_toplevel and env.v.LOG:
            builder.close_stdin()
            builder.start_stdin_log_reader(
                status=True, details=True,
                pretty=True, color=True, debug_locks=False, debug_pids=False)
        else:
            logs.setup(
                tty=sys.stderr, parent_logs=env.v.LOG,
                pretty=env.v.PRETTY, color=env.v.COLOR)
        if env.v.TARGET and not env.v.UNLOCKED:
            me = os.path.join(env.v.STARTDIR,
                              os.path.join(env.v.PWD, env.v.TARGET))
            f = state.File(name=me)
            debug2('TARGET: %r %r %r\n'
                   % (env.v.STARTDIR, env.v.PWD, env.v.TARGET))
        else:
            f = me = None
            debug2('redo-ifchange: not adding depends.\n')
        jobserver.setup(0)
        try:
            if f:
                for t in targets:
                    f.add_dep('m', t)
                f.save()
                state.commit()
            rv = builder.run(targets, should_build)
        finally:
            try:
                state.rollback()
            finally:
                try:
                    jobserver.force_return_tokens()
                except Exception as e:  # pylint: disable=broad-except
                    traceback.print_exc(100, sys.stderr)
                    err('unexpected error: %r\n' % e)
                    rv = 1
    except (KeyboardInterrupt, helpers.ImmediateReturn):
        if env.is_toplevel:
            builder.await_log_reader()
        sys.exit(200)
    state.commit()
    if env.is_toplevel:
        builder.await_log_reader()
    sys.exit(rv)


if __name__ == '__main__':
    main()



"""Code for checking redo target dependencies."""
import os
from . import cycles, env, state
from .logs import debug

CLEAN = 0
DIRTY = 1

def isdirty(f, depth, max_changed,
            already_checked,
            is_checked=state.File.is_checked,
            set_checked=state.File.set_checked_save,
            log_override=state.warn_override):
    """Determine if the given state.File needs to be built.

    Args:
      f: a state.File representing the target to check.
      depth: a string of whitespace representing the recursion depth
        (initially '')
      max_changed: initially the current runid.  If a target is newer than
        this, anything that depends on it is considered outdated.
      already_checked: initially [].  A list of dependencies already
        checked in this recursive cycle, to avoid infinite loops.
      is_checked: a function that returns whether a given state.File has
        already been checked for dirtiness.
      set_checked: a function that marks a given state.File as having now
        been checked for dirtiness.
      log_override: a function that logs a "manual override" warning when
        needed.  (redo-ood replaces this with a no-op.)

    Returns:
      [targets...] if we won't be sure until the given list of targets has
        been built.
      DIRTY if the given target is definitely dirty.
      CLEAN if the given target is definitely not dirty.
    """
    if f.id in already_checked:
        raise cycles.CyclicDependencyError()
    # make a copy of the list, so upon returning, our parent's copy
    # is unaffected
    already_checked = list(already_checked) + [f.id]

    if env.v.DEBUG >= 1:
        debug('%s?%s %r,%r\n'
              % (depth, f.nicename(), f.is_generated, f.is_override))

    if f.failed_runid:
        debug('%s-- DIRTY (failed last time)\n' % depth)
        return DIRTY
    if f.changed_runid is None:
        debug('%s-- DIRTY (never built)\n' % depth)
        return DIRTY
    if f.changed_runid > max_changed:
        debug('%s-- DIRTY (built %d > %d; %d)\n'
              % (depth, f.changed_runid, max_changed, env.v.RUNID))
        return DIRTY  # has been built more recently than parent
    if is_checked(f):
        if env.v.DEBUG >= 1:
            debug('%s-- CLEAN (checked)\n' % depth)
        return CLEAN  # has already been checked during this session
    if not f.stamp:
        debug('%s-- DIRTY (no stamp)\n' % depth)
        return DIRTY

    newstamp = f.read_stamp()
    if f.stamp != newstamp:
        if newstamp == state.STAMP_MISSING:
            debug('%s-- DIRTY (missing)\n' % depth)
            if f.stamp and f.is_generated:
                # previously was stamped and generated, but suddenly missing.
                # We can safely forget that it is/was a target; if someone
                # does redo-ifchange on it and it doesn't exist, we'll mark
                # it a target again, but if someone creates it by hand,
                # it'll be a source.  This should reduce false alarms when
                # files change from targets to sources as a project evolves.
                debug('%s   converted target -> source %r\n' % (depth, f.id))
                f.is_generated = f.failed_runid = 0
                f.save()
                f.refresh()
                assert not f.is_generated
        else:
            debug('%s-- DIRTY (mtime)\n' % depth)
        if f.csum:
            return [f]
        else:
            return DIRTY

    must_build = []
    for mode, f2 in f.deps():
        dirty = CLEAN
        if mode == 'c':
            if os.path.exists(os.path.join(env.v.BASE, f2.name)):
                debug('%s-- DIRTY (created)\n' % depth)
                dirty = DIRTY
        elif mode == 'm':
            sub = isdirty(f2, depth=depth + '  ',
                          max_changed=max(f.changed_runid,
                                          f.checked_runid or 0),
                          already_checked=already_checked,
                          is_checked=is_checked,
                          set_checked=set_checked,
                          log_override=log_override)
            if sub:
                debug('%s-- DIRTY (sub)\n' % depth)
                dirty = sub
        else:
            assert mode in ('c', 'm')
        if not f.csum:
            # f is a "normal" target: dirty f2 means f is instantly dirty
            if dirty == DIRTY:
                # f2 is definitely dirty, so f definitely needs to
                # redo.
                return DIRTY
            elif isinstance(dirty, list):
                # our child f2 might be dirty, but it's not sure yet.  It's
                # given us a list of targets we have to redo in order to
                # be sure.
                must_build += dirty
        else:
            # f is "checksummable": dirty f2 means f needs to redo,
            # but f might turn out to be clean after that (ie. our parent
            # might not be dirty).
            if dirty == DIRTY:
                # f2 is definitely dirty, so f definitely needs to
                # redo.  However, after that, f might turn out to be
                # unchanged.
                return [f]
            elif isinstance(dirty, list):
                # our child f2 might be dirty, but it's not sure yet.  It's
                # given us a list of targets we have to redo in order to
                # be sure.
                must_build += dirty

    if must_build:
        # f is *maybe* dirty because at least one of its children is maybe
        # dirty.  must_build has accumulated a list of "topmost" uncertain
        # objects in the tree.  If we build all those, we can then
        # redo-ifchange f and it won't have any uncertainty next time.
        return must_build
    debug('%s-- CLEAN\n' % (depth,))

    # if we get here, it's because the target is clean
    if f.is_override:
        log_override(f.name)
    set_checked(f)
    return CLEAN



"""redo-log: print past build logs. """
from __future__ import print_function
import errno, fcntl, os, re, struct, sys, time
import termios
from .atoi import atoi
from . import env, logs, options, state

optspec = """
redo-log [options...] [targets...]
--
r,recursive     show build logs for dependencies too
u,unchanged     show lines for dependencies not needing to be rebuilt
f,follow        keep watching for more lines to be appended (like tail -f)
no-details      only show 'redo' recursion trace, not build output
no-status       don't display build summary line in --follow
no-pretty       don't pretty-print logs, show raw @@REDO output instead
no-color        disable ANSI color; --color to force enable (default: auto)
debug-locks     print messages about file locking (useful for debugging)
debug-pids      print process ids in log messages (useful for debugging)
ack-fd=         (internal use only) print REDO-OK to this fd upon starting
"""
o = options.Options(optspec)
(opt, flags, extra) = o.parse(sys.argv[1:])
targets = extra

topdir = os.getcwd()
already = set()
depth = []
total_lines = 0
status = None
start_time = time.time()


# regexp for matching "redo" lines in the log, which we use for recursion.
# format:
#            redo  path/to/target which might have spaces
#            redo  [unchanged] path/to/target which might have spaces
#            redo  path/to/target which might have spaces (comment)
REDO_LINE_RE = re.compile(r'^@@REDO:([^@]+)@@ (.*)\n$')


def _tty_width():
    s = struct.pack("HHHH", 0, 0, 0, 0)
    try:
        s = fcntl.ioctl(sys.stderr.fileno(), termios.TIOCGWINSZ, s)
    except (IOError, ImportError):
        return atoi(os.environ.get('WIDTH')) or 70
    (ysize, xsize, ypix, xpix) = struct.unpack('HHHH', s)
    return xsize or 70


def is_locked(fid):
    return (fid is not None) and not state.Lock(fid=fid).trylock()


def _fix_depth():
    env.v.DEPTH = len(depth) * '  '


def _rel(top, mydir, path):
    return os.path.relpath(os.path.join(top, mydir, path), topdir)


def catlog(t):
    """Copy the given log content to our current log output device.

    Note: this function's behaviour depends on global command-line options.
    """
    global total_lines, status
    lines_written = 0
    interrupted = 0
    if t in already:
        return 0
    if t != '-':
        depth.append(t)
    _fix_depth()
    already.add(t)
    mydir = os.path.dirname(t)
    if t == '-':
        f = sys.stdin
        fid = None
        loglock = None
        logname = None
    else:
        try:
            sf = state.File(name=t, allow_add=False)
        except KeyError:
            sys.stderr.write('redo-log: [%s] %r: not known to redo.\n'
                             % (os.getcwd(), t,))
            sys.exit(24)
        fid = sf.id
        del sf
        state.rollback()
        logname = state.logname(fid)
        loglock = state.Lock(fid + state.LOG_LOCK_MAGIC)
        loglock.waitlock(shared=True)
        f = None
    delay = 0.01
    was_locked = is_locked(fid)
    line_head = ''
    width = _tty_width()
    while 1:
        if not f:
            try:
                f = open(logname)
            except IOError as e:
                if e.errno == errno.ENOENT:
                    # ignore files without logs
                    pass
                else:
                    raise
        if f:
            # Note: normally includes trailing \n.
            # In 'follow' mode, might get a line with no trailing \n
            # (eg. when ./configure is halfway through a test), which we
            # deal with below.
            line = f.readline()
        else:
            line = None
        if not line and (not opt.follow or not was_locked):
            # file not locked, and no new lines: done
            break
        if not line:
            was_locked = is_locked(fid)
            if opt.follow:
                # Don't display status line for extremely short-lived runs
                if opt.status and time.time() - start_time > 1.0:
                    width = _tty_width()
                    head = 'redo %s ' % ('{:,}'.format(total_lines))
                    tail = ''
                    for n in reversed(depth):
                        remain = width - len(head) - len(tail)
                        # always leave room for a final '... ' prefix
                        if remain < len(n) + 4 + 1 or remain <= 4:
                            if len(n) < 6 or remain < 6 + 1 + 4:
                                tail = '... %s' % tail
                            else:
                                start = len(n) - (remain - 3 - 1)
                                tail = '...%s %s' % (n[start:], tail)
                            break
                        elif n != '-':
                            tail = n + ' ' + tail
                    status = head + tail
                    if len(status) > width:
                        sys.stderr.write(
                            '\nOVERSIZE STATUS (%d):\n%r\n'
                            % (len(status), status))
                    assert len(status) <= width
                    sys.stdout.flush()
                    sys.stderr.write('\r%-*.*s\r' % (width, width, status))
                time.sleep(min(delay, 1.0))
                delay += 0.01
            continue
        total_lines += 1
        delay = 0.01
        if not line.endswith('\n'):
            line_head += line
            continue
        if line_head:
            line = line_head + line
            line_head = ''
        if status:
            sys.stdout.flush()
            sys.stderr.write('\r%-*.*s\r' % (width, width, ''))
            status = None
        g = re.match(REDO_LINE_RE, line)
        if g:
            # FIXME: print prefix if @@REDO is not at start of line.
            #   logs.PrettyLog does it, but only if we actually call .write().
            words, text = g.groups()
            kind, pid, when = words.split(':')[0:3]
            pid = atoi(pid)
            relname = _rel(topdir, mydir, text)
            fixname = os.path.normpath(os.path.join(mydir, text))
            if kind == 'unchanged':
                if opt.unchanged:
                    if opt.debug_locks:
                        logs.meta(kind, relname, pid=pid)
                    elif fixname not in already:
                        logs.meta('do', relname, pid=pid)
                    if opt.recursive:
                        if loglock:
                            loglock.unlock()
                        got = catlog(os.path.join(mydir, text))
                        interrupted += got
                        lines_written += got
                        if loglock:
                            loglock.waitlock(shared=True)
                    already.add(fixname)
            elif kind in ('do', 'waiting', 'locked', 'unlocked'):
                if opt.debug_locks:
                    logs.meta(kind, relname, pid=pid)
                    logs.write(line.rstrip())
                    interrupted += 1
                    lines_written += 1
                elif fixname not in already:
                    logs.meta('do', relname, pid=pid)
                    interrupted += 1
                    lines_written += 1
                if opt.recursive:
                    assert text
                    if loglock:
                        loglock.unlock()
                    got = catlog(os.path.join(mydir, text))
                    interrupted += got
                    lines_written += got
                    if loglock:
                        loglock.waitlock(shared=True)
                already.add(fixname)
            elif kind == 'done':
                rv, name = text.split(' ', 1)
                logs.meta(kind, rv + ' ' + _rel(topdir, mydir, name))
                lines_written += 1
            else:
                logs.write(line.rstrip())
                lines_written += 1
        else:
            if opt.details:
                if interrupted:
                    d = env.v.DEPTH
                    env.v.DEPTH = env.v.DEPTH[:-2]
                    logs.meta('resumed', t)
                    env.v.DEPTH = d
                    interrupted = 0
                logs.write(line.rstrip())
                lines_written += 1
    if loglock:
        loglock.unlock()
    if status:
        sys.stdout.flush()
        sys.stderr.write('\r%-*.*s\r' % (width, width, ''))
        status = None
    if line_head:
        # partial line never got terminated
        print(line_head)
    if t != '-':
        assert depth[-1] == t
        depth.pop(-1)
    _fix_depth()
    return lines_written


def main():
    queue = []
    try:
        if not targets:
            sys.stderr.write(
                'redo-log: give at least one target; ' +
                'maybe "all"?\n')
            sys.exit(1)
        state.init(targets)
        if opt.status < 2 and not os.isatty(2):
            opt.status = False
        logs.setup(
            tty=sys.stdout, parent_logs=False,
            pretty=opt.pretty, color=opt.color)
        if opt.debug_locks:
            env.v.DEBUG_LOCKS = 1
        if opt.debug_pids:
            env.v.DEBUG_PIDS = 1
        if opt.ack_fd:
            # Write back to owner, to let them know we started up okay and
            # will be able to see their error output, so it's okay to close
            # their old stderr.
            ack_fd = int(opt.ack_fd)
            assert ack_fd > 2
            if os.write(ack_fd, b'REDO-OK\n') != 8:
                raise Exception('write to ack_fd returned wrong length')
            os.close(ack_fd)
        queue += targets
        while queue:
            t = queue.pop(0)
            if t != '-':
                logs.meta('do', _rel(topdir, '.', t), pid=0)
            catlog(t)
    except KeyboardInterrupt:
        sys.exit(200)
    except IOError as e:
        if e.errno == errno.EPIPE:
            # this happens for example if calling `redo-log | head`, so stdout
            # is piped into another program that closes the pipe before reading
            # all our output.
            # from https://docs.python.org/3/library/signal.html#note-on-sigpipe:
            # Python flushes standard streams on exit; redirect remaining output
            # to devnull to avoid another BrokenPipeError at shutdown
            devnull = os.open(os.devnull, os.O_WRONLY)
            os.dup2(devnull, sys.stdout.fileno())
            sys.exit(141)  # =128+13: "Terminated by SIGPIPE (signal 13)"
        else:
            raise


if __name__ == '__main__':
    main()
